/***************************************************************************
 *   Copyright (C) 2006 by Gerd v. Egidy                                   *
 *   gve@intra2net.com                                                     *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License version   *
 *   2.1 as published by the Free Software Foundation.                     *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this program; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <sstream>

#include "server.hxx"
#include "log.hxx"

namespace libt2n
{

server_connection::server_connection(int _timeout)
    : connection()
{
    set_timeout(_timeout);
    reset_timeout();
    connection_id=0;
    my_server=0;
}


/// get pointer to logging stream, returns NULL if no logging needed
std::ostream* server_connection::get_logstream(log_level_values level)
{
    if (my_server != NULL)
        return my_server->get_logstream(level);
}

/// check if timeout is expired, close connection if so
void server_connection::check_timeout()
{
    if (timeout != -1 && last_action_time+timeout < time(NULL))
    {
        LOGSTREAM(debug,"timeout on connection " << connection_id << ", closing");
        this->close();
    }
}

/// reset the timeout, e.g. if something is received
void server_connection::reset_timeout()
{
    last_action_time=time(NULL);
}

server::server()
{
    set_default_timeout(30);
    set_logging(NULL,none);
    next_id=1;
}

server::~server()
{
    std::map<unsigned int, server_connection*>::iterator ie=connections.end();
    for(std::map<unsigned int, server_connection*>::iterator i=connections.begin(); i != ie; i++)
        delete i->second;
}

int server::add_connection(server_connection* newconn)
{
    unsigned int cid=next_id++;
    newconn->set_id(cid);
    newconn->set_server(this);
    connections[cid]=newconn;

    LOGSTREAM(debug,"new connection accepted, id: " << cid);

    return cid;
}

/// activate logging to the given stream. everything above the given level is logged.
void server::set_logging(std::ostream *_logstream, log_level_values _log_level)
{
    log_level=_log_level;
    logstream=_logstream;
}

/**
    @brief Gets a connection by id
    
    @param conn_id Connection ID
    
    @retval Pointer to connection object
*/
server_connection* server::get_connection(unsigned int conn_id)
{
    std::map<unsigned int, server_connection*>::iterator p=connections.find(conn_id);
    if (p==connections.end())
        return NULL;
    else
        return p->second;
}

/// check for timeouts, remove closed connections. don't forget to call this from time to time.
void server::cleanup()
{
    std::map<unsigned int, server_connection*>::iterator ie=connections.end();
    for(std::map<unsigned int, server_connection*>::iterator i=connections.begin(); i != ie; i++)
        i->second->check_timeout();

    for(std::map<unsigned int, server_connection*>::iterator i=connections.begin(); i != ie;)
    {
        if (i->second->is_closed() && !i->second->packet_available())
        {
            // closed and no usable data in buffer -> remove
            LOGSTREAM(debug,"removing conneciton " << i->first << " because it is closed and no more data waiting");

            delete i->second;
            connections.erase(i);
            i=connections.begin();
            ie=connections.end();
        }
        else
            i++;
    }
}

/** @brief get a complete data packet from any client. The packet is removed from the
            connection buffer.
    @param[out] data the data package
    @param[out] conn_id the connection id we got this packet from
    @retval true if packet found
*/
bool server::get_packet(std::string& data, unsigned int& conn_id)
{
    // todo: this is somehow unfair: the first connections in the map get checked more
    // often than the others and can thus block them out

    std::map<unsigned int, server_connection*>::iterator ie=connections.end();
    for(std::map<unsigned int, server_connection*>::iterator i=connections.begin(); i != ie; i++)
        if (i->second->get_packet(data))
        {
            LOGSTREAM(debug,"got packet (" << data.size() << " bytes) from connection " << i->first);

            conn_id=i->first;
            return true;
        }

    return false;
}

/// get pointer to logging stream, returns NULL if no logging needed
std::ostream* server::get_logstream(log_level_values level)
{
    if (logstream && level >= log_level)
        return logstream;
}
};
